"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Any, Dict, Optional, Union

from qilib.utils.storage.interface import (NoDataAtKeyError,
                                           NodeAlreadyExistsError,
                                           StorageInterface, NodeDoesNotExistsError)
from qilib.utils.type_aliases import TagType


class StorageMemory(StorageInterface):
    """ Reference implementation of StorageInterface with in-memory backend.

    Implements a storage tree as an in-memory dictionary.
    """

    class __Leaf:
        def __init__(self, data: Any) -> None:
            self.data: Any = data

    class __Node(Dict[str, Any]):
        pass

    def __init__(self, name: str) -> None:
        """ In memory implementation of storage class.

        See also: `StorageInterface`

        """
        super().__init__(name)
        self._data: Dict[str, Any] = {}

    @staticmethod
    def _retrieve_value_from_dict_by_tag(dictionary: Dict[str, Any], tag: TagType,
                                         field: Optional[Union[str, int]] = None) -> Any:
        if len(tag) == 0:
            if not isinstance(dictionary, StorageMemory.__Leaf):
                raise NoDataAtKeyError()
            if field is None:
                return dictionary.data
            else:
                if field in dictionary.data:
                    return dictionary.data[field]
                else:
                    raise NoDataAtKeyError(f'The field {field} does not exist.')

        tag_prefix: str = tag[0]
        if tag_prefix not in dictionary:
            raise NoDataAtKeyError(tag)
        return StorageMemory._retrieve_value_from_dict_by_tag(dictionary[tag_prefix], tag[1:], field)

    @staticmethod
    def _retrieve_nodes_from_dict_by_tag(dictionary: Dict[str, Any],
                                         tag: TagType, limit: int = 0) -> TagType:
        if len(tag) == 0:
            results = list(dictionary.keys())
            if limit > 0:
                results = results[:limit]
            return results
        tag_prefix = tag[0]
        if tag_prefix not in dictionary:
            raise NoDataAtKeyError(tag)
        if isinstance(dictionary[tag_prefix], StorageMemory.__Leaf):
            raise NoDataAtKeyError(tag)

        return StorageMemory._retrieve_nodes_from_dict_by_tag(dictionary[tag_prefix], tag[1:], limit=limit)

    @staticmethod
    def _store_value_to_dict_by_tag(dictionary: Dict[str, Any], tag: TagType, value: Any,
                                    field: Optional[Union[str, int]] = None) -> None:
        if len(tag) == 1:
            if tag[0] in dictionary:
                if isinstance(dictionary[tag[0]], StorageMemory.__Node):
                    raise NodeAlreadyExistsError()
            if field is None:
                dictionary[tag[0]] = StorageMemory.__Leaf(value)
            else:
                if tag[0] in dictionary:
                    dictionary[tag[0]].data[field] = value
                else:
                    raise NodeDoesNotExistsError(f'The Node {tag[0]} does not exist')
        else:
            if not tag[0] in dictionary:
                dictionary[tag[0]] = StorageMemory.__Node()
            elif isinstance(dictionary[tag[0]], StorageMemory.__Leaf):
                raise NodeAlreadyExistsError(f'Cannot store or replace data, because \'{tag[0]}\' is already a node')

            StorageMemory._store_value_to_dict_by_tag(dictionary[tag[0]], tag[1:], value, field)

    def load_data(self, tag: TagType) -> Any:
        if not isinstance(tag, list):
            raise TypeError()
        return self._unserialize(StorageMemory._retrieve_value_from_dict_by_tag(self._data, tag))

    def save_data(self, data: Any, tag: TagType) -> None:
        self._validate_tag(tag)
        StorageMemory._store_value_to_dict_by_tag(self._data, tag, self._serialize(data))

    def get_latest_subtag(self, tag: TagType) -> Optional[TagType]:
        child_tags: TagType = self.list_data_subtags(tag)
        if len(child_tags) == 0:
            return None
        child_tags = sorted(child_tags)
        return tag + [child_tags[-1]]

    def list_data_subtags(self, tag: TagType, limit: int = 0) -> TagType:
        try:
            tags: TagType = self._retrieve_nodes_from_dict_by_tag(self._data, tag, limit=limit)
        except NoDataAtKeyError:
            return []
        return list(tags)

    def search(self, query: str) -> Any:
        raise NotImplementedError()

    def tag_in_storage(self, tag: TagType) -> bool:
        tmp_data = self._data
        for partial_tag in tag:
            if partial_tag in tmp_data:
                tmp_data = tmp_data[partial_tag]
            else:
                return False
        return True

    def load_individual_data(self, tag: TagType, field: Union[str, int]) -> Any:
        """ Retrieve the value of an individual field at the given tag

        Args:
            tag: The tag
            field: Name of the field to load

        Returns:
            Value of the field

        """
        self._validate_tag(tag)
        self._validate_field(field)
        return self._unserialize(StorageMemory._retrieve_value_from_dict_by_tag(self._data, tag, field))

    def update_individual_data(self, data: Any, tag: TagType, field: Union[str, int]) -> None:
        """ Update the value of an individual field at the given tag.
        If the field does not already exist, then it will be created.

         Args:
             data: The data to be used for updating the field
             tag: The tag
             field: Name of the field to update

         """
        self._validate_tag(tag)
        self._validate_field(field)
        StorageMemory._store_value_to_dict_by_tag(self._data, tag, self._serialize(data), field)



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import inspect
from types import ModuleType
from typing import Dict, Tuple, Optional, Type, Any

from qilib.configuration_helper.instrument_adapter import InstrumentAdapter


class InstrumentAdapterFactory:
    """ Creates InstrumentAdapters with corresponding QCoDeS driver.

    Factory class responsible for creating instances of the InstrumentAdapter and the corresponding
    QCoDeS instrument. It is called with the name of the desired InstrumentAdapter sub-class for the
    instrument as well as a dictionary that provides key-value pairs of address specifications.
    """

    #: A container that holds all instantiated instrument adapters created by the factory.
    #: When an adapter is requested via the get_instrument_adapter method, this dictionary is
    #: checked first for an existing adapter with same name and address.
    adapter_instances: Dict[Tuple[str, str], InstrumentAdapter] = {}

    #: A dictionary containing all available adapters. External adapters can be added via add_instrument_adapter_package
    #: or add_instrument_adapter_class.
    _instrument_adapters: Dict[str, Type[InstrumentAdapter]] = {}

    @classmethod
    def add_instrument_adapter_package(cls, package: ModuleType) -> None:
        """ Adds InstrumentAdapters (from an external package)

        Args:
            package: The package that contains InstrumentAdapters

        """

        for adapter in inspect.getmembers(package, lambda m: inspect.isclass(m) and issubclass(m, InstrumentAdapter)):
            cls.add_instrument_adapter_class(adapter[1])

    @classmethod
    def add_instrument_adapter_class(cls, adapter: Type[InstrumentAdapter]) -> None:
        """
        Args:
            adapter: Adapter class to be added to the factory.
        Raises:
            TypeError: If the adapter is not subclass of InstrumentAdapter.

        """
        if not issubclass(adapter, InstrumentAdapter):
            raise TypeError(f'{adapter.__name__} is not a subclass of InstrumentAdapter')
        cls._instrument_adapters[adapter.__name__] = adapter

    @classmethod
    def get_instrument_adapter(cls, instrument_adapter_class_name: str, address: str,
                               instrument_name: Optional[str] = None) -> InstrumentAdapter:
        """ Factory method for creating an InstrumentAdapter with QCoDeS instrument.

        Args:
            instrument_adapter_class_name: Name of the InstrumentAdapter subclass.
            address: Address of the physical instrument.
            instrument_name: An optional name for the underlying instrument.

        Returns:
             An instance of the requested InstrumentAdapter.

        Raises:
            ValueError: If adapter is not found or adapter exist but with a different instrument name.

        """
        instrument_adapter_key = instrument_adapter_class_name, str(address)
        if instrument_adapter_key in cls.adapter_instances:
            return cls._get_adapter_instance(instrument_adapter_key, instrument_name)
        elif instrument_adapter_class_name in cls._instrument_adapters:
            return cls._get_new_adapter_instance(address, instrument_adapter_class_name, instrument_name)
        else:
            raise ValueError(f"No such InstrumentAdapter {instrument_adapter_class_name}")

    @classmethod
    def _get_new_adapter_instance(cls, address: str, class_name: str,
                                  instrument_name: Optional[str] = None) -> InstrumentAdapter:

        args: Any = (address, instrument_name) if instrument_name is not None else (address, )
        adapter: InstrumentAdapter = cls._instrument_adapters[class_name](*args)
        cls.adapter_instances[(class_name, str(address))] = adapter
        return adapter

    @classmethod
    def _get_adapter_instance(cls, instrument_adapter_key: Tuple[str, str],
                              instrument_name: Optional[str] = None) -> InstrumentAdapter:

        adapter = cls.adapter_instances[instrument_adapter_key]
        if instrument_name is not None and adapter.instrument is not None \
           and instrument_name != adapter.instrument.name:
            raise ValueError(f'An adapter exist with different instrument name'
                             f' \'{adapter.instrument.name}\' != \'{instrument_name}\'')
        return adapter



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import List, Optional

from qilib.configuration_helper.instrument_adapter_factory import InstrumentAdapterFactory
from qilib.configuration_helper.exceptions import DuplicateTagError
from qilib.configuration_helper.visitor import Visitor
from qilib.utils.python_json_structure import PythonJsonStructure
from qilib.utils.storage.interface import StorageInterface


class InstrumentConfiguration:
    """ Associates a configuration with an InstrumentAdapter and allows it to be stored or retrieved from storage."""

    STORAGE_BASE_TAG = 'configuration'

    def __init__(self, adapter_class_name: str, address: str, storage: StorageInterface,
                 tag: Optional[List[str]] = None, configuration: Optional[PythonJsonStructure] = None,
                 instrument_name: Optional[str] = None) -> None:
        """ A set of instrument configurations

        Args
            instrument_adapter_class_name: Name of the InstrumentAdapter subclass
            address: Address of the physical instrument
            storage: Any storage that implements the StorageInterface
            tag: A unique identifier for a instrument configuration set
            configuration: The instrument configuration
            instrument_name: User defined name for the instrument
        """
        self._adapter_class_name = adapter_class_name
        self._address = address
        self._storage = storage
        self._instrument_name = instrument_name
        self._adapter = InstrumentAdapterFactory.get_instrument_adapter(adapter_class_name, address, instrument_name)
        self._configuration = PythonJsonStructure() if configuration is None else configuration
        self._tag = [self.STORAGE_BASE_TAG, adapter_class_name, StorageInterface.datetag_part()] if tag is None else tag

    def __repr__(self) -> str:
        repr_string = f'{self.__class__.__name__}({self._adapter_class_name!r}, {self._address!r}, {self._storage!r}, '\
            f'{self._tag!r}, {self._configuration!r}, {self._instrument_name!r})'
        return repr_string

    @property
    def tag(self) -> List[str]:
        """ A unique identifier for this instrument configuration set """
        return self._tag

    @property
    def storage(self) -> StorageInterface:
        """ The storage interface used """
        return self._storage

    @property
    def address(self) -> str:
        """ The address of the physical instrument """
        return self._address

    @property
    def configuration(self) -> PythonJsonStructure:
        """ The instrument configuration """
        return self._configuration

    @staticmethod
    def load(tag: List[str], storage: StorageInterface) -> 'InstrumentConfiguration':
        """ A factory that creates a new InstrumentConfiguration by loading from database.

        Args:
            tag: A unique identifier for a instrument configuration.
            storage: Default mongo database, but can optionally be any storage that implements the StorageInterface.

        Returns:
            A new InstrumentConfiguration loaded from database.

        """
        document = storage.load_data(tag)
        adapter_class_name = document['adapter_class_name']
        address = document['address']
        instrument_name = document.get('instrument_name', None)
        configuration = document['configuration']
        return InstrumentConfiguration(adapter_class_name, address, storage, tag, configuration, instrument_name)

    def store(self) -> None:
        """ Saves object to storage.

         Raises:
             DuplicateTagError: If this object's tag is already in the database.

        """
        if self._storage.tag_in_storage(self._tag):
            raise DuplicateTagError(
                f"InstrumentConfiguration for {self._adapter_class_name} with tag '{self._tag}' already in storage")
        document = PythonJsonStructure(adapter_class_name=self._adapter_class_name,
                                       address=self._address,
                                       instrument_name=self._instrument_name,
                                       configuration=self._configuration)
        self._storage.save_data(document, self._tag)

    def apply(self) -> None:
        """ Uploads the configuration to the instrument."""
        self._adapter.apply(self._configuration)

    def apply_delta(self, update: bool = True) -> None:
        """ Compare configuration with instrument and apply configuration that differs.

        Args:
            update: If True, request all parameter values from instrument, else use latest set values.

        """
        instrument_config = self._adapter.read(update=update)
        delta = self._get_configuration_delta(instrument_config)
        self._adapter.apply(delta)

    def _get_configuration_delta(self, instrument_config: PythonJsonStructure) -> PythonJsonStructure:
        delta = PythonJsonStructure()
        for parameter, configuration in self._configuration.items():
            if 'value' in configuration and configuration['value'] != instrument_config[parameter]['value']:
                delta[parameter] = configuration
        return delta

    def apply_delta_lazy(self) -> None:
        """ Compare configuration with instrument driver last known settings and apply configuration that differs."""
        self.apply_delta(update=False)

    def refresh(self) -> None:
        """ Read the settings from the instrument and updated configuration.

        If the settings read from the instrument differs from the configuration the tag is also updated.
        """
        instrument_config = self._adapter.read(update=True)
        delta = self._get_configuration_delta(instrument_config)
        if len(delta) > 0 or len(instrument_config) != len(self._configuration):
            self._configuration = instrument_config
            self._tag = [self.STORAGE_BASE_TAG, self._adapter_class_name, StorageInterface.datetag_part()]

    def accept(self, visitor: Visitor) -> None:
        """ Accept a visitor, run visit method with self as a parameter and propagate to instrument adapter.

        Args:
            visitor: An implementation of the Visitor interface.

        """
        visitor.visit(self)
        self._adapter.accept(visitor)



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Optional, Dict, Any, Union, MutableMapping

from bson.objectid import ObjectId
from pymongo.change_stream import CollectionChangeStream
from pymongo.errors import DuplicateKeyError
from pymongo.mongo_client import MongoClient
from qilib.data_set.data_array import DataArray
from qilib.utils.serialization import NumpyArrayEncDec
from qilib.utils.type_aliases import EncodedNumpyArray, NumpyNdarrayType


class DocumentNotFoundError(Exception):
    """ Error that is raised when document is not found."""


class FieldNotUniqueError(Exception):
    """ Error raised if field is not unique"""


class MongoDataSetIO:
    """ Helper class for the MongoDataSetIOReader and -Writer"""

    DEFAULT_DATABASE_NAME = 'qilib'
    DEFAULT_COLLECTION_NAME = 'data_sets'

    def __init__(self, name: Optional[str] = None, document_id: Optional[str] = None,
                 create_if_not_found: Optional[bool] = True, database: str = DEFAULT_DATABASE_NAME,
                 collection: str = DEFAULT_COLLECTION_NAME) -> None:
        """

        Args:
            name: Name of the document.
            document_id: The document _id.
            create_if_not_found: Create a new document if no match is found.
            database: Name of the database.
            collection: Name of the collections.

        Raises:
            DocumentNotFoundError: If document not found in database.
        """

        if name is None and document_id is None:
            raise DocumentNotFoundError("Neither 'name' nor 'document_id' were provided.")

        self._client = MongoClient()  # type: MongoClient[Any]
        self._db = self._client[database][collection]
        self._assert_name_field_is_unique()

        query_dict: MutableMapping[str, Any] = {}
        if name is not None:
            query_dict['name'] = name
        if document_id is not None:
            query_dict['_id'] = ObjectId(document_id)
        document = self._db.find_one(query_dict)
        if document is None:
            if name is not None and document_id is None and create_if_not_found:
                self._id = str(self._db.insert_one(query_dict).inserted_id)
                self._name = name
            else:
                raise DocumentNotFoundError("Document not found in database.")
        else:
            self._id = str(document.get('_id'))
            self._name = document.get('name')

    @property
    def name(self) -> str:
        return self._name

    @property
    def id(self) -> str:
        return self._id

    def watch(self) -> CollectionChangeStream[Any]:
        """ Start watching the underlying document for updates.

        Returns:
            A blocking watch-cursor iterator that returns updates when available.

        """
        pipeline = [{'$match': {'fullDocument.name': self._name}}]
        cursor = self._db.watch(pipeline=pipeline, full_document='updateLookup')
        return cursor

    def get_document(self) -> Any:
        """ Get the complete document from the database.

        Returns:
            A complete document.

        """
        document = self._db.find_one({"_id": ObjectId(self._id)})
        return document

    def finalize(self) -> None:
        """ Close the connection to the database."""
        self._client.close()

    def append_to_document(self, data: Dict[str, Any]) -> None:
        """ Append data to an array in the underlying document.

        Args:
            data: Data to append to the document.

        """
        self._db.update_one(
            {"name": self._name},
            {"$push": data,
             "$currentDate": {"lastModified": True}})

    def update_document(self, data: Dict[str, Any]) -> None:
        """ Update data in the underlying document.

        Args:
            data: Data to be updated.

        """
        self._db.update_one(
            {"name": self._name},
            {"$set": data,
             "$currentDate": {"lastModified": True}})

    @staticmethod
    def encode_numpy_array(array: Union[NumpyNdarrayType, DataArray]) -> EncodedNumpyArray:
        """ Encode numpy array to store in database.
        Args:
            array: Numpy array to encode.

        Returns:
            The encoded array.

        """
        if isinstance(array, DataArray):
            array = array.data
        return NumpyArrayEncDec.encode(array)

    @staticmethod
    def decode_numpy_array(encoded_array: Dict[str, Any]) -> NumpyNdarrayType:
        """ Decode a numpy array from database.

        Args:
            encoded_array: The encoded array to decode.

        Returns:
            The decoded array.
        """
        return NumpyArrayEncDec.decode(encoded_array)

    def _assert_name_field_is_unique(self) -> None:
        """ The field 'name' should be unique in the database.

        Raises:
            FieldNotUniqueError: If there is already a duplicate name in the database or if this method otherwise
                fails to set 'name' unique.

        """
        try:
            self._db.create_index("name", unique=True)
        except DuplicateKeyError as e:
            raise FieldNotUniqueError("Failed to set field 'name' unique.") from e
        else:
            index_info = self._db.index_information()
            if index_info['name_1']['unique'] is False:
                raise FieldNotUniqueError("Field 'name' is not unique in database.")



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Any, Optional, Type

from qilib.configuration_helper.adapters.common_instrument_adapter import CommonInstrumentAdapter
from qilib.utils.python_json_structure import PythonJsonStructure


class M4iInstrumentAdapter(CommonInstrumentAdapter):
    """ Adapter for the Spectrum M4i digitizer."""

    def __init__(self, address: str, instrument_name: Optional[str] = None,
                 instrument_class: Optional[Type[Any]] = None) -> None:
        from qcodes_contrib_drivers.drivers.Spectrum.M4i import M4i
        super().__init__(address, instrument_name)
        if instrument_class is None:
            instrument_class = M4i
        self._instrument: M4i = instrument_class(self._instrument_name, cardid=address)
        if self._instrument is not None:
            self._instrument.initialize_channels()

    def _filter_parameters(self, parameters: PythonJsonStructure) -> PythonJsonStructure:
        if parameters['box_averages']['value'] == 1:
            parameters.pop('box_averages')

        parameters.pop('card_available_length')
        parameters.pop('IDN')
        return parameters



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from queue import Queue, Empty
from threading import Thread, ThreadError
from typing import Optional, Any, Dict, List

from pymongo.change_stream import CollectionChangeStream
from pymongo.errors import InvalidOperation, OperationFailure

from qilib.data_set.data_set import DataSet
from qilib.data_set.data_array import DataArray
from qilib.data_set.mongo_data_set_io import MongoDataSetIO
from qilib.data_set.data_set_io_reader import DataSetIOReader

watchers = []


class MongoDataSetIOReader(DataSetIOReader):
    """ Allows a DataSet to subscribe to changes, and updates, in a mongodb."""

    THREAD_ERROR = 'thread_error'

    def __init__(self, name: Optional[str] = None, document_id: Optional[str] = None,
                 database: str = MongoDataSetIO.DEFAULT_DATABASE_NAME,
                 collection: str = MongoDataSetIO.DEFAULT_COLLECTION_NAME) -> None:
        """ DataSetIOReader implementation for a mongodb.

        Note:
            The finalize method has to be called to close database connection and join the watcher thread.

        Args:
            name: Name of data set in the underlying mongodb.
            document_id: Name of data set in the underlying mongodb.
            database: Name of the database.
            collection: Name of the collections.
        Raises:
            DocumentNotFoundError: If no data set with document_id or name found in database.

        """

        super().__init__()
        self._name = name
        self._id = document_id
        self._mongo_data_set_io = MongoDataSetIO(name, document_id, create_if_not_found=False, database=database,
                                                 collection=collection)
        self._set_arrays: Dict[str, DataArray] = {}
        self._watcher = self._mongo_data_set_io.watch()
        watchers.append(self._watcher)
        self._update_queue = Queue()  # type: ignore
        self._update_thread = Thread(target=self._update_worker, args=(self._update_queue, self._watcher))
        self._update_thread.daemon = True
        self._update_thread.start()
        self._data_set: Any = None

    def __del__(self) -> None:
        self._watcher.close()
        watchers.remove(self._watcher)
        self._update_thread.join(1)

    def sync_from_storage(self, timeout: float) -> None:
        """ Poll the Mongo database for changes and apply any to the bound data_set.

        Args:
            timeout: Stop syncing if collecting an item takes longer than the timeout time.
                The timeout can be -1 (blocking), 0 (non-blocking), or >0 (wait at most that many seconds).

        Raises:
            TimeoutError: If timeout is reached while the storage queue is still empty

        """
        blocking = timeout != 0
        empty_queue = self._update_queue.empty() if timeout == 0 else False
        while not empty_queue:
            try:
                document = self._update_queue.get(blocking, timeout if timeout > 0 else None)
            except Empty as e:
                raise TimeoutError from e
            if MongoDataSetIOReader.THREAD_ERROR in document:
                raise ThreadError('Watcher thread has stopped unexpectedly.') from document[
                    MongoDataSetIOReader.THREAD_ERROR]
            updated_fields = document['updateDescription']['updatedFields']
            adjusted_updates = self._convert_dot_notation_to_dict(updated_fields)
            self._update_data_set(adjusted_updates)
            empty_queue = self._update_queue.empty()

    def _convert_dot_notation_to_dict(self, updated_fields: Dict[str, Any]) -> Dict[str, Any]:
        """ If a nested field is updated in the mongo database, the nested document is replace with a dot notation that
            does not match the data set that is to be updated.
            For example:
                {data_arrays.array_name: {<data>}}
            Needs to be converted to:
                {data_arrays: {array_name: {<data>}}

        Args:
            updated_fields: An update event from the mongo database change stream.

        Returns:
            An updated dictionary without the dot notation.

        """
        adjusted_updates: Dict[str, Any] = {}
        for update in updated_fields.keys():
            key = update.split('.')
            if len(key) > 1 and key[1].isnumeric():
                # {array_updates.index: [<index>, <data>] -> {array_updates: [[<index>, <data>]]
                adjusted_updates[key[0]] = [updated_fields[update]]
            elif len(key) > 1 and (key[0] == self.DATA_ARRAYS or key[0] == self.METADATA):
                # {data_arrays.array_name: {<data>}} -> {data_arrays: {array_name: {<data>}}
                adjusted_updates[key[0]] = {key[1]: updated_fields[update]}
            else:
                adjusted_updates[key[0]] = updated_fields[update]
        return adjusted_updates

    def bind_data_set(self, data_set: DataSet) -> None:
        """ Binds the DataSet to the DataSetIOReader.

        Args:
            data_set: The object that encompasses DataArrays.

        """
        self._data_set = data_set
        document = self._mongo_data_set_io.get_document()
        self._data_set.name = document.get('name')
        self._update_data_set(document)

    def _update_data_set(self, document: Any) -> None:

        if self.METADATA in document:
            for field, value in document.get(self.METADATA).items():
                setattr(self._data_set, field, value)
        if self.DATA_ARRAYS in document:
            set_arrays = list(filter(lambda a: a['is_setpoint'], document.get(self.DATA_ARRAYS).values()))
            self._update_set_arrays(set_arrays)
            data_arrays = list(filter(lambda a: not a['is_setpoint'], document.get(self.DATA_ARRAYS).values()))
            for array in data_arrays:
                if hasattr(self._data_set, array['name']):
                    self._update_data_array(array)
                else:
                    self._data_set.add_array(self._construct_data_array(array))
        if self.ARRAY_UPDATES in document:
            for array_update in document.get(self.ARRAY_UPDATES):
                index_or_slice = tuple(array_update[0]) if isinstance(array_update[0], list) else array_update[0]
                data = array_update[1]
                self._data_set.add_data(index_or_slice, data)

    @staticmethod
    def load(name: Optional[str] = None, document_id: Optional[str] = None,
             database: str = MongoDataSetIO.DEFAULT_DATABASE_NAME,
             collection: str = MongoDataSetIO.DEFAULT_COLLECTION_NAME) -> DataSet:
        """ Load an existing data set from the mongodb.

        Args:
            name: Name of the data set.
            document_id: _id of the data set.
            database: Name of the database.
            collection: Name of the collections.

        Returns:
            A new instance of the underlying data set.

        Raises:
            DocumentNotFoundError: If document_id or name do not match any data set in database.

        """
        reader = MongoDataSetIOReader(name, document_id, database=database, collection=collection)
        return DataSet(storage_reader=reader)

    def _update_set_arrays(self, arrays: List[Dict[str, Any]]) -> None:
        for array in arrays:
            if array['name'] not in self._set_arrays:
                self._set_arrays[array['name']] = self._construct_data_array(array)

    def _construct_data_array(self, array: Dict[str, Any]) -> DataArray:
        set_arrays = [self._set_arrays[name] for name in array['set_arrays']]
        data_array = DataArray(name=array['name'],
                               label=array['label'],
                               unit=array['unit'],
                               is_setpoint=array['is_setpoint'],
                               preset_data=MongoDataSetIO.decode_numpy_array(array['preset_data']),
                               set_arrays=set_arrays)
        return data_array

    def _update_data_array(self, array: Dict[str, Any]) -> None:
        np_array = MongoDataSetIO.decode_numpy_array(array['preset_data'])
        self._data_set.data_arrays[array['name']].label = array['label']
        self._data_set.data_arrays[array['name']].unit = array['unit']

        for i in range(len(np_array)):
            self._data_set.data_arrays[array['name']][i] = np_array[i]

    @staticmethod
    def _update_worker(queue: Any, watcher: CollectionChangeStream[Any]) -> None:
        try:
            while True:
                document = watcher.next()
                queue.put(document)
        except (StopIteration, InvalidOperation, OperationFailure) as e:
            queue.put({MongoDataSetIOReader.THREAD_ERROR: e})



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC, abstractmethod
from typing import Any, Union, Dict, Tuple

from qilib.data_set.data_array import DataArray


class DataSetIOWriter(ABC):
    """ Abstract base class for data set io writers."""

    def __init__(self) -> None:
        self._finalized = False

    @abstractmethod
    def sync_metadata_to_storage(self, field_name: str, value: Any) -> None:
        """ Registers a change to a metadata field.

        This function creates a new field if the field name does not exists.

        Args:
            field_name: The metadata field name.
            value: The metadata value to change.
        """

    @abstractmethod
    def sync_data_to_storage(self, index_or_slice: Union[int, Tuple[int]], data: Dict[str, Any]) -> None:
        """ Registers a DataArray update to the DataSetIO.

        Args:
            index_or_slice: The indices of the DataArray to update.
            data: Name of the DataArray to be updated and the new value.
        """

    @abstractmethod
    def sync_add_data_array_to_storage(self, data_array: DataArray) -> None:
        """ Registers a new DataArray event.

        Args:
            data_array: A container for measurement data and setpoint arrays.
        """

    @abstractmethod
    def finalize(self) -> None:
        """ Sets the data IO to read-only.

            No more data will be written after applying this function and triggers the closing
            of file handles, or optimizes event streams.
        """

    def _is_finalized(self) -> None:
        if self._finalized:
            raise ValueError('Operation on closed IO writer.')



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from typing import Tuple, Type, Optional, cast
from spirack import SPI_rack
from qilib.configuration_helper.serial_port_resolver import SerialPortResolver
from qilib.configuration_helper.adapters.spi_rack_instrument_adapter import SpiRackInstrumentAdapter
from qilib.configuration_helper.adapters.common_instrument_adapter import CommonInstrumentAdapter

from qilib.utils.python_json_structure import PythonJsonStructure


class SpiModuleInstrumentAdapter(CommonInstrumentAdapter):

    def __init__(self, address: str, instrument_name: Optional[str] = None,
                 spi_rack_instrument_adapter_class: Optional[Type[SpiRackInstrumentAdapter]] = None) -> None:
        """ The base instrument adapter for each SPI rack module.

        Each module is identified using the name of the spirack the module number. The address is
        the combined value, e.g. 'spirack1_module10'.

        Args:
            address: The unique identifier of the SPI rack module, which should be of the form
                     <spi_name>_module<module_number>, where spi_name is the name of the spi rack
                     and module_number the integer module number.
            instrument_name: An optional name for the underlying instrument.
        """
        super().__init__(address, instrument_name)
        identifier, self._module_number = SpiModuleInstrumentAdapter.__collect_settings(address)
        if spi_rack_instrument_adapter_class is None:
            spi_rack_instrument_adapter_class = SpiRackInstrumentAdapter
        adapter: SpiRackInstrumentAdapter = cast(SpiRackInstrumentAdapter,
                                                 SerialPortResolver.get_serial_port_adapter(
                                                     spi_rack_instrument_adapter_class.__name__, identifier))
        self._spi_rack: SPI_rack = adapter.spi_rack

    def read(self, update: bool = True) -> PythonJsonStructure:
        """ Reads and returns all SPI rack module settings from the device.

        All module parameters will be collected, even if the parameter is write only.
        If the write only parameter has not been set, a None will be set as value.

        Args:
            update: Update the settings first before returning the settings configuration.

        Returns:
            A custom dictionary with all the SPI rack module settings.
        """
        return PythonJsonStructure(super().read(update))

    def _filter_parameters(self, parameters: PythonJsonStructure) -> PythonJsonStructure:
        """ Removes the identifier parameter from the configuration settings.

        Args:
            parameters: A custom dictionary with all the SPI rack module settings.
        """
        parameters.pop('IDN')
        return parameters

    @staticmethod
    def __collect_settings(address: str) -> Tuple[str, int]:
        """ Collects the SPI rack name and the module number from the address identifier.

        Args:
            address: The address of the spirack and the module.

        Returns:
            Contains two items, the spirack identifier as string and the module number as integer.
        """
        results = re.match(r'(\w+)_module(\d+)', address)
        if results is not None:
            identifier = results.group(1)
            module = int(results.group(2))
        else:
            error_message = "Invalid address format. ({} != <spi_name>_module<module_number>)".format(address)
            raise ValueError(error_message)
        return identifier, module



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import math
from typing import Any, Optional, Type

from qcodes.instrument_drivers.american_magnetics.AMI430 import AMI430

from qilib.configuration_helper.adapters.read_only_configuration_instrument_adapter import \
    ReadOnlyConfigurationInstrumentAdapter
from qilib.utils.python_json_structure import PythonJsonStructure


class AMI430InstrumentAdapter(ReadOnlyConfigurationInstrumentAdapter):
    """ Adapter for the AMI430 QCoDeS driver."""

    def __init__(self, address: str, instrument_name: Optional[str] = None,
                 instrument_class: Optional[Type[AMI430]] = None) -> None:
        """ Instantiate a new AMI430 instrument adapter.

        Args:
            address: IP address and port number separated by a column, 'x.x.x.x:xxxx'
            instrument_name: An optional name for the underlying instrument.
        """
        super().__init__(address)
        ip_and_port = address.split(':')
        name = instrument_name if instrument_name is not None else self.name
        if instrument_class is None:
            instrument_class = AMI430
        self._instrument: AMI430 = instrument_class(name=name, address=ip_and_port[0], port=int(ip_and_port[1]))
        self.field_variation_tolerance = 0.01

    def apply(self, config: PythonJsonStructure) -> None:
        """ Does not apply config only  compares config to device settings.

        Configuration Parameter with setter commands are matched for equality, on mismatch
        error is raised

        Args:
            config: Containing the instrument configuration.
        """
        super().apply(config)

    def _filter_parameters(self, parameters: PythonJsonStructure) -> PythonJsonStructure:
        for values in parameters.values():
            if 'val_mapping' in values:
                values.pop('val_mapping')
        return parameters

    def _compare_config_values(self, config_value: Any, device_value: Any, parameter: Optional[str] = None) -> bool:
        if parameter == 'field':
            delta = math.fabs(config_value - device_value)
            return delta > self.field_variation_tolerance
        else:
            return bool(config_value != device_value)



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import logging
from abc import ABC, abstractmethod
from typing import Optional

from qcodes.instrument.base import Instrument

from qilib.configuration_helper.visitor import Visitor
from qilib.utils.python_json_structure import PythonJsonStructure


class InstrumentAdapter(ABC):

    def __init__(self, address: str, instrument_name: Optional[str] = None) -> None:
        """ A template for an adapter to the QCoDeS instrument interface.

        Args:
            address: The address/ID of the QCoDeS instrument.
        """
        self._name = f'{self.__class__.__name__}_{address}'
        self._address = address
        self._instrument: Optional[Instrument] = None
        self._instrument_name = instrument_name if instrument_name is not None else self.name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._address!r}, {self._instrument_name!r})'

    @property
    def name(self) -> str:
        return self._name

    @property
    def instrument(self) -> Optional[Instrument]:
        return self._instrument

    @property
    def address(self) -> str:
        return self._address

    @abstractmethod
    def apply(self, config: PythonJsonStructure) -> None:
        """ An abstract method that should be implemented in subclasses.

        Writes a full set of configuration to the adapters instrument.

        Args:
            config: Containing the instrument configuration.
        """

    def read(self, update: bool = False) -> PythonJsonStructure:
        """ Obtains a full set of settings from the instrument.

        Returns:
            Part of the instrument snapshot, i.e., parameter values, without the instrument's
            parameters which are explicitly filtered out, and the instrument name.
        """
        configuration = PythonJsonStructure()
        if self._instrument is not None:
            snapshot = self._instrument.snapshot(update)
            parameters = self._filter_parameters(snapshot['parameters'])
            valued_parameters = self.__notify_and_remove_none_values(parameters)
            configuration.update(valued_parameters)

        return configuration

    @abstractmethod
    def _filter_parameters(self, parameters: PythonJsonStructure) -> PythonJsonStructure:
        """ Filters out parameters that are not used for instrument configuration storage.

        This function should be overwritten in the subclasses for each specific instrument,
        if needed when reading the configuration.

        Args:
            parameters: A complete snapshot from an instrument.

        Returns:
            PythonJsonStructure: Contains the instrument snapshot without the instrument
                                 parameters which are filtered out by this function.
        """

    def close_instrument(self) -> None:
        """ Close the wrapped QCoDeS instrument."""
        if self._instrument is not None:
            self._instrument.close()

    def accept(self, visitor: Visitor) -> None:
        """ Accept a visitor and run visit method with self as a parameter.

        Args:
            visitor: An implementation of the Visitor interface.
        """
        visitor.visit(self)

    def __notify_and_remove_none_values(self, parameters: PythonJsonStructure) -> PythonJsonStructure:
        """ Return parameters from the QCoDeS snapshot which are not None.

            Takes the parameters of the QCoDeS instrument snapshot. Removes all parameters
            which have a value of None. Returns the parameter settings which have a value.
            All parameters with None value will be listed in the log as an error.

        Args:
            parameters: The parameters of a QCoDeS instrument snapshot.

        Returns:
            PythonJsonStructure: Contains the instrument snapshot parameters without the
                                 instrument parameters which a none value.
        """
        valued_parameters = PythonJsonStructure()
        none_value_parameters = PythonJsonStructure()
        for parameter_name, settings in parameters.items():
            if 'value' in settings and settings['value'] is None:
                none_value_parameters[parameter_name] = settings
            else:
                valued_parameters[parameter_name] = settings

        if none_value_parameters:
            parameter_names = list(none_value_parameters.keys())
            error_message = f'Parameter values of {self._instrument_name} are None. Please set: {parameter_names}.'
            logging.error(error_message)

        return valued_parameters



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from copy import deepcopy
from typing import cast, Optional, Tuple, List, Union, Any, Set, Dict

import numpy
import numpy as np

from qilib.utils.type_aliases import NumpyNdarrayType


class DataArray:
    """ A container for measurement data and setpoint arrays.

    All attributes/methods not defined in the DataArray are delegated to an underlying numpy array.

    The DataArray makes sure that the dimensions of the set arrays are correct with regards to the data array and vice
    versa (e.g., trying to set a 1D array of 10 elements as the data array with a 1D setpoint array of 8 elements will
    raise an error).
    """

    def __init__(self, name: str, label: str, unit: str = '', is_setpoint: bool = False,
                 preset_data: Optional[NumpyNdarrayType] = None,
                 set_arrays: Optional[Union[List['DataArray'], Tuple['DataArray', ...]]] = None,
                 shape: Optional[Tuple[int, ...]] = None) -> None:
        """
        Args:
            name:  Name for the data array
            label: Label, e.g. x-axis if it is a setpoint.
            unit: Unit for the measurement, or setpoint data.
            is_setpoint: If the DataArray is a setpoint.
            preset_data: Initialize the DataArray with predefined data. Note that a copy of the numpy
                array is stored in data, not the actual array.
            set_arrays: a list of setpoint arrays.
            shape: Initialize with a shape rather than predefined data.
        """

        self._name: str = name
        self._label: str = label
        self._unit: str = unit
        self._is_setpoint: bool = is_setpoint
        self._data: NumpyNdarrayType
        if preset_data is not None:
            self._data = np.array(preset_data).copy()
        elif shape is not None:
            self._data = np.ndarray(shape) * np.NAN
        else:
            raise TypeError("Required arguments 'shape' or 'preset_data' not found")
        self._set_arrays = set_arrays if set_arrays is not None else []
        if len(self._set_arrays) > 0:
            self._verify_array_dimensions()

    def __add__(self, other: Union[float, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__add__(other)

    def __mul__(self, other: Union[float, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__mul__(other)

    def __matmul__(self, other: Union[List[Union[float, int]], 'DataArray', NumpyNdarrayType]) -> numpy.int64:
        return cast(numpy.int64, self._data.__matmul__(other))

    def __pow__(self, other: Union[float, int]) -> NumpyNdarrayType:
        return self._data.__pow__(other)

    def __sub__(self, other: Union[float, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__sub__(other)

    def __mod__(self, other: Union[float, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__mod__(other)

    def __floordiv__(self, other: Union[float, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__floordiv__(other)

    def __lshift__(self, other: Union[int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__lshift__(other)

    def __rshift__(self, other: Union[int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__rshift__(other)

    def __and__(self, other: Union[bool, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__and__(other)

    def __or__(self, other: Union[bool, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__or__(other)

    def __xor__(self, other: Union[bool, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__xor__(other)

    def __truediv__(self, other: Union[float, int, 'DataArray', NumpyNdarrayType]) -> NumpyNdarrayType:
        return self._data.__truediv__(other)

    def __getattr__(self, name: str) -> Any:
        if hasattr(np.ndarray, name):
            return getattr(self._data, name)
        raise AttributeError("DataArray has no attribute '{}'".format(name))

    def __dir__(self) -> Set[str]:
        data_dir = dir(self._data)
        data_array_dir = list(super(DataArray, self).__dir__())
        return set(data_dir + data_array_dir)

    def __str__(self) -> str:
        array_names = [array.name for array in self.set_arrays] if self.set_arrays is not None else []
        print_string = "{} {}: {}\ndata: {}\nset_arrays:{}".format(self.__class__.__name__, self.shape, self.name,
                                                                   self._data, array_names)
        return print_string

    def __repr__(self) -> str:
        return "DataArray(id=%r, name=%r, label=%r, unit=%r, is_setpoint=%r, data=%r, set_arrays=%r)" % (
            id(self), self._name, self._label, self._unit, self._is_setpoint, self._data, self._set_arrays)

    def __getitem__(self, index: Union[Tuple[int, ...], int]) -> Any:
        return self._data[index]

    def __setitem__(self, index: Union[Tuple[int, ...], int], data: Any) -> None:
        self._data[index] = data

    def __copy__(self) -> 'DataArray':
        data_array_copy = type(self)(name=self._name, label=self._label, shape=self.shape)
        data_array_copy.__dict__.update(self.__dict__)
        return data_array_copy

    def __deepcopy__(self, memo: Dict[int, Any]) -> 'DataArray':
        data_array_copy = type(self)(name=self._name, label=self._label, shape=self.shape)
        data_array_copy.__dict__.update(self.__dict__)
        data_array_copy._data = deepcopy(self._data, memo)
        data_array_copy._set_arrays = deepcopy(self._set_arrays, memo)

        return data_array_copy

    def __len__(self) -> int:
        return len(self._data)

    def _verify_array_dimensions(self) -> None:
        shapes = [array.shape for array in self._set_arrays]
        shapes.sort(key=lambda s: len(s))
        if self.is_setpoint:
            shapes.append(self._data.shape)
        else:
            if shapes[-1] != self.shape:
                raise ValueError("Dimensions of 'set_arrays' and 'data' do not match.")
        DataArray._check_dimensions(shapes)

    @staticmethod
    def _check_dimensions(shapes: List[Tuple[int]]) -> None:
        for i in range(len(shapes)):
            dim = shapes.pop(0)
            if not all(shape[i] == dim[i] for shape in shapes):
                raise ValueError("Dimensions of 'set_arrays' do not match.")

    @property
    def data(self) -> NumpyNdarrayType:
        return self._data

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def unit(self) -> str:
        return self._unit

    @unit.setter
    def unit(self, unit: str) -> None:
        self._unit = unit

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, label: str) -> None:
        self._label = label

    @property
    def is_setpoint(self) -> bool:
        return self._is_setpoint

    @property
    def set_arrays(self) -> Union[List['DataArray'], Tuple['DataArray', ...]]:
        return self._set_arrays



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Dict, Any, Optional, cast

import numpy as np

from qilib.utils.type_aliases import PJSValues


class PythonJsonStructure(dict):  # type: ignore
    __serializable_key_types = (str, int)
    __serializable_python_types = (bool, int, float, complex, str, bytes)
    __serializable_numpy_types = (np.float32, np.float64, np.int32, np.int64, np.cfloat)
    __serializable_value_types = (*__serializable_python_types, *__serializable_numpy_types)

    def __init__(self, *args: Dict[str, PJSValues], **kwargs: Any) -> None:
        """ A python container which can hold data objects and can be serialized
            into JSON. Currently the following data types can be added to the
            container object: bool, bytes, int, float, list, tuple, dict
            PythonJsonStructure and numpy arrays. The PythonJsonStructure is a
            dictionary with similar calling methods, but with some
            limitations/restrictions.

        Args:
            *args: A serializable dictionary with key value pair data.
            **kwargs: Arbitrary keyword arguments with serializable values.
        """
        super().__init__()
        self.update(*args, **kwargs)

    def __setitem__(self, key: str, value: PJSValues) -> None:
        """ Appends or changes an item of the container.

        Args:
            key: The key of the container item.
            value: The value of the updated item.
        """

        value = self.__validate_key_value_pair(key, value)
        super().__setitem__(key, value)

    def update(self, *args: Any, **kwargs: Any) -> None:
        """ Update the PythonJsonStructure with the key/value pairs from other
            dict/PythonJsonStructure, overwriting existing keys."""
        if args:
            args_data = args[0]
            for key, value in args_data.items():
                self[key] = value
        if kwargs:
            kwargs_items = kwargs.items()
            for key, value in kwargs_items:
                self[key] = value

    def setdefault(self, key: str, default: Optional[PJSValues] = None) -> PJSValues:
        """ If key is in the dictionary, return its value. If not, insert key
            with a value of default and return default.

        Args:
            key: The key of the container item.
            default: The default value of the updated item.
        """
        default = self.__validate_key_value_pair(key, default)
        return cast(PJSValues, super().setdefault(key, default))

    def __validate_key_value_pair(self, key: str, value: Optional[PJSValues] = None) -> PJSValues:
        if isinstance(value, dict):
            value = PythonJsonStructure(value)
        PythonJsonStructure.__assert_correct_key_type(key)
        self.__check_serializability(value)
        return value

    def __check_serializability(self, data: Any) -> None:
        if isinstance(data, (list, tuple)):
            for item in data:
                self.__check_serializability(item)

        elif isinstance(data, dict):
            for key, item in data.items():
                PythonJsonStructure.__assert_correct_key_type(key)
                self.__check_serializability(item)

        elif data is not None:
            PythonJsonStructure.__is_valid_type(data)

    @staticmethod
    def __assert_correct_key_type(key: Any) -> None:
        if not isinstance(key, PythonJsonStructure.__serializable_key_types):
            raise TypeError('Invalid key! (key={})'.format(key))

    @staticmethod
    def __is_valid_type(data: Any) -> None:
        if isinstance(data, np.ndarray):
            valid = data.dtype in PythonJsonStructure.__serializable_numpy_types
            data_type = data.dtype
        else:
            valid = any([isinstance(data, type) for type in PythonJsonStructure.__serializable_value_types])
            data_type = type(data)
        if not valid:
            raise TypeError('Data is not serializable ({})!'.format(data_type))



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC, abstractmethod
from typing import List

from qilib.configuration_helper.instrument_adapter import InstrumentAdapter
from qilib.utils.python_json_structure import PythonJsonStructure


class CommonInstrumentAdapter(InstrumentAdapter, ABC):

    def apply(self, config: PythonJsonStructure) -> None:
        """ Applies the given instrument configuration settings onto the adapters instrument.

        Only the setter commands will be updated.
        Note that setter only parameters which have not been set yield a None when reading the configuration from the
        instrument adapter. The apply method will raise a ValueError if any of the configuration parameter values are
        None. The None value parameters are listed in the raised ValueError.

        Args:
            config: The configuration with settings for the adapters instrument.
        """
        if self._instrument is not None:
            parameters = []
            for parameter in config:
                if parameter in self._instrument.parameters and hasattr(self._instrument.parameters[parameter], 'set'):
                    parameters.append(parameter)

            self.__raise_none_value_parameters(config, parameters)
            for parameter in parameters:
                self._instrument.set(parameter, config[parameter]['value'])

    @abstractmethod
    def _filter_parameters(self, parameters: PythonJsonStructure) -> PythonJsonStructure:
        """ Filters out parameters that are not used for instrument configuration storage.

        This function should be overwritten in the subclasses for each specific instrument,
        if needed when reading the configuration.

        Args:
            parameters: A complete snapshot from an instrument.

        Returns:
            PythonJsonStructure: Contains the instrument snapshot without the instrument
                                 parameters which are filtered out by this function.
        """

    def __raise_none_value_parameters(self, config: PythonJsonStructure, parameters: List[PythonJsonStructure]) -> None:
        """ Raises a ValueError when one or more parameters in the configuration have value None.

        Args:
            config: A configuration with settings for the instrument adapter.
            parameters: A list with settable instrument parameters.
        """
        if self._instrument is not None:
            none_value_parameters = list(filter(lambda parameter: config[parameter]['value'] is None, parameters))
            if none_value_parameters:
                error_message = f'The following parameter(s) of {self._instrument.name} have value ' \
                                f'None and cannot be set: {none_value_parameters}!'
                raise ValueError(error_message)



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Any, Optional
from abc import ABC, abstractmethod

from qilib.configuration_helper.instrument_adapter import InstrumentAdapter
from qilib.utils.python_json_structure import PythonJsonStructure


class ConfigurationError(Exception):
    """ Error to raise if configuration does not match."""


class ReadOnlyConfigurationInstrumentAdapter(InstrumentAdapter, ABC):
    """ Common methods to verify value mismatch on config apply"""

    def apply(self, config: PythonJsonStructure) -> None:
        """ Does comparison for config values with set command.

        Args:
            config: The configuration with settings for the adapters instrument.

        Raises:
            ConfigurationError: If config does not match device configuration .
        """
        if self._instrument is not None:
            device_config = self.read(True)
            for parameter in config:
                if parameter in self._instrument.parameters and hasattr(self._instrument.parameters[parameter], 'set'):
                    result = self._compare_config_values(config[parameter]['value'],
                                                         device_config[parameter]['value'], parameter)
                    if result:
                        self._raise_configuration_error(config[parameter]['value'], device_config[parameter]['value'],
                                                        parameter)

    @abstractmethod
    def _compare_config_values(self, config_value: Any, device_value: Any, parameter: Optional[str] = None) -> bool:
        """ Comparison logic for  configuration parameter values.
        This function should be overwritten in the subclasses for each specific instrument.

        Args:
            config_value: Configuration value as supplied as argument to apply
            device_value: Configuration value as read from the device
            parameter: Name of the configuration parameter with setter command (Optional)

        Returns:
            True or False based on the comparison result
        """

    @staticmethod
    def _raise_configuration_error(config_value: Any, device_value: Any, parameter: str) -> None:
        raise ConfigurationError(
            "Configuration for {} does not match: '{}' != '{}'".format(parameter, config_value, device_value))



from qilib.configuration_helper.instrument_adapter import InstrumentAdapter
from qilib.configuration_helper.instrument_adapter_factory import InstrumentAdapterFactory
from qilib.configuration_helper.instrument_configuration import InstrumentConfiguration
from qilib.configuration_helper.instrument_configuration_set import InstrumentConfigurationSet
from qilib.configuration_helper.instrument_configuration_visitor import InstrumentConfigurationVisitor
from qilib.configuration_helper.serial_port_resolver import SerialPortResolver
from qilib.configuration_helper.configuration_helper import ConfigurationHelper
from qilib.configuration_helper import adapters

InstrumentAdapterFactory.add_instrument_adapter_package(adapters)



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from queue import Empty

from qilib.data_set.data_set import DataSet
from qilib.data_set.data_set_io_reader import DataSetIOReader
from qilib.utils.memory_storage_queue import MemoryStorageQueue


class MemoryDataSetIOReader(DataSetIOReader):
    """ Allows a DataSet to subscribe to changes, and updates, in an in-memory data storage."""

    def __init__(self, storage_queue: MemoryStorageQueue) -> None:
        """ Construct a new instance of MemoryDataSetIOReader.
            This should not be called directly but a Reader/Writer pair should be created with
            the MemoryDataSetIOFactory.

        Args:
            storage_queue: Fifo shared with a MemoryDataSetIOWriter.
        """
        super().__init__()
        self._storage_queue = storage_queue

    def sync_from_storage(self, timeout: float) -> None:
        """ Poll the MemoryStorageQueue for changes and apply any to the bound data_set.

          Args:
              timeout: Stop syncing if collecting an item takes more then a the timeout time.
                       The timeout can be -1 (blocking), 0 (non-blocking), or >0 (wait at most that many seconds).

          Raises:
                TimeoutError: If timeout is reached while the storage queue is still empty
        """

        blocking = timeout != 0
        empty_queue = self._storage_queue.empty() if timeout == 0 else False
        while not empty_queue:
            try:
                data_type, storage_data = self._storage_queue.get(blocking, timeout if timeout > 0 else None)
            except Empty as e:
                raise TimeoutError from e

            if data_type == self.DATA_ARRAY:
                self._data_set.add_array(storage_data)
            elif data_type == self.DATA:
                self._data_set.add_data(*storage_data)
            elif data_type == self.METADATA:
                setattr(self._data_set, *storage_data)

            empty_queue = self._storage_queue.empty()

    @staticmethod
    def load() -> DataSet:
        """ MemoryDataSetReader has only a fifo storage queue and can therefor not load a
            DataSet from storage.
        """
        raise NotImplementedError('The load function cannot be used with the MemoryDataSetIOReader!')



"""Quantum Inspire library

Copyright 2022 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from operator import itemgetter
from typing import Any, Optional, Union, cast

import numpy as np
from bson.codec_options import TypeCodec, CodecOptions, TypeRegistry
from bson.objectid import ObjectId
from pymongo.mongo_client import MongoClient
from pymongo.errors import ServerSelectionTimeoutError

from qilib.data_set.mongo_data_set_io import MongoDataSetIO
from qilib.utils.serialization import NumpyKeys, Serializer, serializer as _serializer
from qilib.utils.storage.interface import (NoDataAtKeyError,
                                           NodeAlreadyExistsError,
                                           StorageInterface,
                                           ConnectionTimeoutError)
from qilib.utils.type_aliases import TagType


class NumpyArrayCodec(TypeCodec):

    @property
    def python_type(self) -> Any:
        return np.ndarray

    def transform_python(self, value: Any) -> Any:
        return MongoDataSetIO.encode_numpy_array(value)

    @property
    def bson_type(self) -> Any:
        return dict

    def transform_bson(self, value: Any) -> Any:
        if NumpyKeys.OBJECT in value and value[NumpyKeys.OBJECT] == np.array.__name__:
            return MongoDataSetIO.decode_numpy_array(value)

        return value


class StorageMongoDb(StorageInterface):
    """Reference implementation of StorageInterface with an mongodb backend

    Implements a storage tree in a MongoDB collection
    Performance Note: Creating index(es) in mongodb will help improve query performance
    eg: An index on (parent, tag) will improve the performance of queries in _retrieve_nodes_by_tag
    """

    def __init__(self, name: str, host: str = 'localhost', port: int = 27017, database: str = '',
                 serializer: Union[Serializer, None] = None, connection_timeout: float = 30000) -> None:
        """MongoDB implementation of storage class

        See also: `StorageInterface`

        Args:
            name: Symbolic name for the storage instance.
            host: MongoDB host
            port: MongoDB port
            database: The database to use, if empty the name of the storage is used
            connection_timeout: How long to try to connect to database before raising an error in milliseconds
        Raises:
            StorageTimeoutError: If connection to database has not been established before connection_timeout is reached
        """
        super().__init__(name)

        type_registry = TypeRegistry([NumpyArrayCodec()])
        codec_options = CodecOptions(type_registry=type_registry)  # type: CodecOptions[Any]
        self._client = MongoClient(host, port, serverSelectionTimeoutMS=connection_timeout)  # type: MongoClient[Any]
        self._check_server_connection(connection_timeout)
        self._db = self._client.get_database(database or name, codec_options=codec_options)
        self._collection = self._db.get_collection('storage')

        if serializer is None:
            serializer = _serializer
        self._serialize = serializer.encode_data
        self._unserialize = serializer.decode_data

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} at 0x{id(self):x}: name {self._db.name}>'

    def __str__(self) -> str:
        return f'<{type(self).__name__}: name {self._db.name}>'

    def _check_server_connection(self, timeout: float) -> None:
        """ Check if connection has been established to database server."""
        try:
            self._client.server_info()
        except ServerSelectionTimeoutError as e:
            raise ConnectionTimeoutError(f'Failed to connect to Mongo database within {timeout} milliseconds') from e

    def _get_root(self) -> ObjectId:
        """Get or create a root node if it doesn't exist yet

        Returns:
            An ObjectID of the root node
        """

        node = self._collection.find_one({'tag': '', 'parent': {'$exists': False}})

        if node is not None:
            return cast(ObjectId, node['_id'])

        return cast(ObjectId, self._collection.insert_one({'tag': ''}).inserted_id)

    def _retrieve_nodes_by_tag(self, tag: TagType, parent: ObjectId, document_limit: int = 0) -> TagType:
        """Traverse the tree and list the children of a given tag

        Args:
            tag: The node tag
            parent: The ObjectID of the node's parent
            document_limit: Maximum number of documents to return. If set to zero there is no limit

        Returns:
            A list of names of the children
        """
        if len(tag) == 0:
            return list(map(itemgetter('tag'),
                            self._collection.find({'parent': parent, 'tag': {'$exists': True}},
                                                  {'value': 0},
                                                  limit=document_limit,
                                                  sort=[('tag', -1)])))

        else:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0]})
            if doc is None:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" cannot be found')
            elif 'value' in doc:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" is not a node')
            else:
                return self._retrieve_nodes_by_tag(tag[1:], doc['_id'], document_limit)

    def _retrieve_value_by_tag(self, tag: TagType, parent: ObjectId, field: Optional[Union[str, int]] = None) -> Any:
        """Traverse the tree and retrieves the value / field value of a given leaf tag
        If the field is specified, it returns the value of the field instead of tag value
        If the specified field does not exist, it will raise a NoDataAtKeyError

        Args:
            tag: The leaf tag
            parent: The ObjectID of the leaf's parent
            field: The field to be retrieved. Default value is none.

        Returns:
            Data held by the leaf. If field is provided, returns the value of the
            field stored in the leaf

        Raises:
            NoDataAtKeyError: If a tag in Tag list does not exist
            or if the field does not exist
        """

        if len(tag) == 1:
            if field is None:
                doc = self._collection.find_one({'parent': parent, 'tag': tag[0]})
            else:
                doc = self._collection.find_one({'parent': parent, 'tag': tag[0]}, {f'value.{field}': 1})

            if doc is None:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" cannot be found')
            elif 'value' not in doc:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" is not a leaf')
            else:
                if field is None:
                    return doc['value']
                elif field in doc['value']:
                    return doc['value'][field]
                else:
                    raise NoDataAtKeyError(f'The field "{field}" does not exists')

        else:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0], 'value': {'$exists': False}})
            if doc is None:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" cannot be found')
            else:
                return self._retrieve_value_by_tag(tag[1:], doc['_id'], field)

    def _store_value_by_tag(self, tag: TagType, data: Any, parent: ObjectId,
                            field: Optional[Union[str, int]] = None) -> None:
        """ Store a value at a given tag. In case a field is specified, function will update the value of the
        field with data. If the field does not already exists, the field will be created

        Args:
            tag: The tag
            data: Data to store
            parent: An ObjectID of the node's parent
            field: Field to be updated. Default value is None

        Raises:
              NodeAlreadyExistsError: If a tag in Tag List is an unexpected node/leaf
              NoDataAtKeyError:  If a tag in Tag List does not exist
        """

        if len(tag) == 1:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0]})
            if doc:
                if 'value' not in doc:
                    raise NodeAlreadyExistsError(f'Tag "{tag[0]}" is not a leaf')
                elif field is None:
                    self._collection.update_one({'parent': parent, 'tag': tag[0]},
                                                {'$set': {'value': data}})
                else:
                    self._collection.update_one({'parent': parent,
                                                 'tag': tag[0]},
                                                {'$set': {'value.' +
                                                          str(field): data}})
            elif field is None:
                self._collection.insert_one({'parent': parent, 'tag': tag[0], 'value': data})
            else:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" does not exist')
        else:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0]})
            if doc is None:
                if field is None:
                    parent = self._collection.insert_one({'parent': parent, 'tag': tag[0]}).inserted_id
                else:
                    raise NoDataAtKeyError(f'Tag "{tag[0]}" does not exist')
            else:
                if 'value' in doc:
                    raise NodeAlreadyExistsError(f'Tag "{tag[0]}" is a leaf')
                else:
                    parent = doc['_id']

            self._store_value_by_tag(tag[1:], data, parent, field)

    def load_data(self, tag: TagType) -> Any:
        if not isinstance(tag, list):
            raise TypeError('Tag should be a list of strings')

        if len(tag) == 0:
            raise NoDataAtKeyError('Tag cannot be empty')

        return self._unserialize(self._decode_data(self._retrieve_value_by_tag(tag, self._get_root())))

    def save_data(self, data: Any, tag: TagType) -> None:
        self._validate_tag(tag)
        self._store_value_by_tag(tag, self._encode_data(self._serialize(data)), self._get_root())

    def load_individual_data(self, tag: TagType, field: Union[str, int]) -> Any:
        """ Retrieve an individual field value at a given tag

            Args:
                tag: The tag
                field: Name of the individual field to be retrieved

            Raises:
                NoDataAtKeyError if the tag is empty

            Returns:
                Value of the field
        """

        self._validate_tag(tag)
        self._validate_field(field)

        if len(tag) == 0:
            raise NoDataAtKeyError('Tag cannot be empty')

        return self._unserialize(self._decode_data(
            self._retrieve_value_by_tag(tag, self._get_root(), self._encode_field(self._serialize(field)))))

    def update_individual_data(self, data: Any, tag: TagType, field: Union[str, int]) -> None:
        """ Update an individual field at a given tag with the given data.
        If the field does not exist, it will be created.

            Args:
                data: Data to update
                tag: The tag
                field: Name of the individual field to updated

        """

        self._validate_tag(tag)
        self._validate_field(field)
        self._store_value_by_tag(tag, self._encode_data(self._serialize(data)), self._get_root(),
                                 self._encode_field(self._serialize(field)))

    def get_latest_subtag(self, tag: TagType) -> Optional[TagType]:
        """ Get the latest subtag

        Args:
            tag: Tag to search from
        Returns:
            Latest subtag found among subtags or None if there are no subtags
        """
        child_tags = self.list_data_subtags(tag, limit=1)
        if len(child_tags) == 0:
            return None

        return tag + [child_tags[0]]

    def list_data_subtags(self, tag: TagType, limit: int = 0) -> TagType:
        """ List subtags for the given tag. The number of subtags listed is based on the limit parameter (0 for
        listing all subtags)

        Args:
            tag: Tag to search from
            limit: Maximum number of subtags to return. If set to zero there is no limit
        Returns:
            List of subtags found, sorted in descending order
        """
        try:
            tags = self._retrieve_nodes_by_tag(tag, self._get_root(), limit)
        except NoDataAtKeyError:
            tags = []

        return tags

    def search(self, query: str) -> Any:
        raise NotImplementedError()

    def tag_in_storage(self, tag: TagType) -> bool:
        parent = self._get_root()
        for tag_part in tag:
            doc = self._collection.find_one({'parent': parent, 'tag': tag_part}, {'value': 0})
            if doc is None:
                return False
            else:
                parent = doc['_id']
        return True

    @staticmethod
    def _encode_field(field: Union[int, str]) -> str:
        """Encodes a field value. A field can be an individual key in a structure like dict which is to be
        read or updated

        Args:
            field: An integer or a string

        Returns:
            Encoded value for the field in string format

        """

        return StorageMongoDb._encode_int(field) if isinstance(field, int) \
            else StorageMongoDb._encode_str(field)

    @staticmethod
    def _encode_int(value: int) -> str:
        """Encodes an int value to a string

        Args:
            value: An integer

        Returns:
            An encoded integer
        """

        return f'_integer[{value}]'

    @staticmethod
    def _is_encoded_int(value: str) -> bool:
        """Checks if the given value is a properly encoded integer

        Args:
            value: The encoded value

        Returns:
            True if an integer can be decoded, False otherwise
        """

        if not value.startswith('_integer[') or not value.endswith(']'):
            return False

        value = value[len('_integer['):-1]
        return value[1:].isdigit() if value.startswith('-') else value.isdigit()

    @staticmethod
    def _decode_int(value: str) -> int:
        """Parses an integer from the encoded value

        Args:
            value: The encoded integer

        Returns:
            The decoded integer
        """

        if not StorageMongoDb._is_encoded_int(value):
            raise ValueError()

        return int(value[len('_integer['):-1])

    @staticmethod
    def _encode_str(value: str) -> str:
        """Encodes a (dotted) string
        Args:
            value: The value

        Returns:
            The encoded value
        """

        return value.replace('.', '\\u002e')

    @staticmethod
    def _decode_str(value: str) -> str:
        """Decodes a (dotted) string

        Args:
            value: The value

        Returns:
            The decoded value
        """

        return value.replace('\\u002e', '.')

    @staticmethod
    def _encode_data(data: Any) -> Any:
        """Recursively encode the data and apply dot replacement and integer encoding on the keys

        Args:
            data: The data
        Returns:
            The transformed data
        """

        if isinstance(data, dict):
            return {
                StorageMongoDb._encode_str(StorageMongoDb._encode_int(key) if isinstance(key, int) else key)
                : StorageMongoDb._encode_data(value)
                for key, value in data.items()
            }

        elif isinstance(data, list):
            return [StorageMongoDb._encode_data(item) for item in data]

        return data

    @staticmethod
    def _decode_data(data: Any) -> Any:
        """Recursively decode the data and apply dot replacement and integer decoding on the keys

        Args:
            data: The data
        Returns:
            The transformed data
        """

        if isinstance(data, dict):
            return {
                StorageMongoDb._decode_int(StorageMongoDb._decode_str(key))
                if StorageMongoDb._is_encoded_int(key) else StorageMongoDb._decode_str(key)
                : StorageMongoDb._decode_data(value)
                for key, value in data.items()
            }

        elif isinstance(data, list):
            return [StorageMongoDb._decode_data(item) for item in data]

        return data



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from collections.abc import Sequence
from datetime import datetime
from typing import Any, Union, Tuple, List, Dict, Optional

from qilib.data_set.data_array import DataArray
from qilib.data_set.type_aliases import DataArrays
from qilib.utils.python_json_structure import PythonJsonStructure


class DataSet:
    """ A DataSet is an object that encompasses DataArrays.

    A DataArray can have another DataArray (or multiple) as setpoint(s).
    A DataSet can have multiple measurement arrays sharing the same setpoints.
    It is an error to have multiple measurement arrays with different setpoints.
    """

    def __init__(self, storage_writer: Any = None, storage_reader: Any = None, name: str = '',
                 time_stamp: Optional[datetime] = None, user_data: Optional[PythonJsonStructure] = None,
                 data_arrays: Optional[DataArrays] = None, set_arrays: Optional[List[DataArray]] = None) -> None:
        """
        Args:
            storage_writer: Object providing a storage backend.
            name: Name for the DataSet.
            time_stamp: Time of the measurement. datetime.now is used if not provided.
            user_data: User meta-data. E.g. instrument snapshot.
            data_arrays: DataArrays used in the experiment/
            set_arrays: None or more setpoint arrays.

        Raises:
            TypeError: If data_arrays are not of type DataArray.
            ValueError: If setpoints are provided and don't match those of the data_arrays or if multiple data_arrays
                are provided with various set_arrays.
        """
        self._finalized = False
        self._storage_writer = storage_writer if storage_writer is not None else []
        if not isinstance(self._storage_writer, Sequence):
            self._storage_writer = [self._storage_writer]
        self._storage_reader = storage_reader
        self._name = name
        self._time_stamp = datetime.now() if time_stamp is None else time_stamp
        self._user_data = user_data
        self._data_arrays: Dict[str, DataArray] = {}
        self._default_array_name = ""
        self._set_arrays: Dict[str, DataArray] = {}
        if set_arrays is not None:
            self._add_set_arrays(set_arrays)
        if data_arrays is not None:
            self._add_data_arrays(data_arrays)
        if self._storage_reader is not None:
            self._storage_reader.bind_data_set(self)

    def __repr__(self) -> str:
        return "DataSet(id=%r, name=%r, storage_writer=%r, storage_reader=%r, time_stamp=%r, user_data=%r, " \
               "data_arrays=%r, set_arrays=%r)" % (
                   id(self), self._name, self._storage_writer, self._storage_reader, self._time_stamp, self._user_data,
                   self._data_arrays, self._set_arrays)

    def __str__(self) -> str:
        heading = "{}: {}".format(self.__class__.__name__, self._name)
        table = [['name', 'label', 'unit', 'shape', 'setpoint']]
        for name, array in self._data_arrays.items():
            table.append([name, array.label, array.unit, str(array.shape), str(array.is_setpoint)])
        for array in self._set_arrays.values():
            table.append([array.name, array.label, array.unit, str(array.shape), str(array.is_setpoint)])
        return heading + self._format_str(table)

    def add_array(self, array: DataArray) -> None:
        """
        Args:
            array (DataArray): Array to be added to the experiment.

        Raises:
            ValueError: If an array with the same name has already been added.
            ValueError: If setpoints of the array do not match already added setpoints.
            SyntaxError: If array name is not a valid identifier.
        """

        self._verify_array_name(array.name)
        if len(self._set_arrays) == 0:
            self._add_set_arrays(array.set_arrays)
        else:
            self._verify_set_points(array.set_arrays)
        self._data_arrays[array.name] = array
        setattr(self, array.name, array)
        self._default_array_name = self._default_array_name or array.name
        for storage in self._storage_writer:
            storage.sync_add_data_array_to_storage(array)

    def add_data(self, index_or_slice: Union[Tuple[int, ...], int], data: Dict[str, Union[List[Any], Any]]) -> None:
        """ Update an underlying DataArray.

        Args:
            index_or_slice: Setpoints, can be an int or tuple of integers.
            data: Key is the name of the array and the dict value is the data.
        """

        for array_name, data_value in data.items():
            if array_name in self._data_arrays:
                self._data_arrays[array_name][index_or_slice] = data_value
            elif array_name in self._set_arrays:
                self._set_arrays[array_name][index_or_slice] = data_value
            else:
                raise LookupError(f'No such array with name \'{array_name}\' in data set')

        for storage in self._storage_writer:
            storage.sync_data_to_storage(index_or_slice, data)

    def sync_from_storage(self, timeout: int) -> None:
        """ Poll the DataSetIO for changes and apply any to the in-memory representation.
            Timeout can be:
                -1 (blocking),
                 0 (non-blocking)
                >0 (wait at most that many seconds)

        Args:
            timeout: Time to wait in seconds.

        Raises:
            TimeoutError: If storage is empty when timeout has passed.

        """
        if self._storage_reader is not None:
            self._storage_reader.sync_from_storage(timeout)

    def save_to_storage(self) -> None:
        """ Save DataSet to underlying storage."""
        for metadata in ['name', 'time_stamp', 'user_data', 'default_array_name', '_finalized']:
            self._add_metadata_to_storage(metadata, getattr(self, metadata))
        for array in self._set_arrays.values():
            self._add_array_to_storage(array)
        for array in self._data_arrays.values():
            self._add_array_to_storage(array)

    def finalize(self) -> None:
        """ Indicates to DataSetIOWriters that no more data will be written."""
        self._finalized = True
        self.save_to_storage()
        for storage in self._storage_writer:
            storage.finalize()

    def add_storage_writer(self, storage_writer: Any) -> None:
        """ Add a new DataSetIOStorageWriter to the DataSet

        Args:
            storage_writer: A new initialized storage writer.

        """
        if self._storage_reader is not None:
            raise ValueError("It is not allowed to have both storage_reader and storage_writer.")
        self._storage_writer.append(storage_writer)
        self.save_to_storage()

    @property
    def is_finalized(self) -> bool:
        return self._finalized

    @property
    def storage(self) -> Any:
        return self._storage_writer or self._storage_reader

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._add_metadata_to_storage('name', name)
        self._name = name

    @property
    def data_arrays(self) -> Dict[str, 'DataArray']:
        return self._data_arrays

    @property
    def set_arrays(self) -> Dict[str, 'DataArray']:
        return self._set_arrays

    @property
    def time_stamp(self) -> datetime:
        return self._time_stamp

    @time_stamp.setter
    def time_stamp(self, time_stamp: datetime) -> None:
        self._time_stamp = time_stamp
        self._add_metadata_to_storage('time_stamp', time_stamp)

    @property
    def user_data(self) -> Union[None, PythonJsonStructure]:
        return self._user_data

    @user_data.setter
    def user_data(self, user_data: PythonJsonStructure) -> None:
        self._user_data = user_data
        self._add_metadata_to_storage('user_data', user_data)

    @property
    def default_array_name(self) -> str:
        return self._default_array_name

    @default_array_name.setter
    def default_array_name(self, default_array_name: str) -> None:
        self._default_array_name = default_array_name
        self._add_metadata_to_storage('default_array_name', default_array_name)

    def _verify_set_points(self, set_arrays: DataArrays) -> None:
        if list(self._set_arrays.values()) != list(set_arrays):
            raise ValueError('Set point arrays do not match.')

    def _add_set_arrays(self, set_arrays: DataArrays) -> None:
        if isinstance(set_arrays, DataArray):
            set_arrays = [set_arrays]
        elif not isinstance(set_arrays, Sequence):
            raise TypeError("'set_arrays' have to be of type 'DataArray', not {}".format(type(set_arrays)))

        for array in set_arrays:
            self._verify_array_name(array.name)
            self._set_arrays[array.name] = array
            setattr(self, array.name, array)
            for storage in self._storage_writer:
                storage.sync_add_data_array_to_storage(array)

    def _verify_array_name(self, name: str) -> None:
        if not isinstance(name, str):
            raise ValueError("Array name has to be string, not {}".format(type(name)))
        if not name.isidentifier():
            raise SyntaxError("'{}' is an invalid name for an identifier.".format(name))
        if hasattr(self, name):
            raise ValueError("DataSet already contains an array with the name '{}'".format(name))

    def _add_data_arrays(self, data_arrays: DataArrays) -> None:
        if isinstance(data_arrays, Sequence):
            for array in data_arrays:
                self._add_data_arrays(array)
        elif isinstance(data_arrays, DataArray):
            self.add_array(data_arrays)
        else:
            raise TypeError("'data_arrays' have to be of type 'DataArray', not {}".format(type(data_arrays)))

    @staticmethod
    def _format_str(table: List[List[str]]) -> str:
        row_template = '\n  {info[0]:{lens[0]}} | {info[1]:{lens[1]}} | {info[2]:{lens[2]}} | {info[3]:{lens[3]}} | ' \
                       '{info[4]}'
        formatted_string = ''
        column_lengths = []
        for i in range(4):
            column_lengths.append(len(max(table, key=lambda t: len(t[i]))[i]))
        for row in table:
            formatted_string += row_template.format(info=row, lens=column_lengths)
        return formatted_string

    def _add_metadata_to_storage(self, field_name: str, value: Any) -> None:
        for storage in self._storage_writer:
            storage.sync_metadata_to_storage(field_name, value)

    def _add_array_to_storage(self, array: DataArray) -> None:
        for storage in self._storage_writer:
            storage.sync_add_data_array_to_storage(array)
